
#include "ns3/core-module.h"
#include "ns3/network-module.h"
#include "ns3/mobility-module.h"
#include "ns3/config-store-module.h"
#include "ns3/wifi-module.h"
#include "ns3/internet-module.h"
#include <fstream>
#include <string>
#include "ns3/point-to-point-module.h"
#include "ns3/applications-module.h"
#include "ns3/flow-monitor-module.h"
#include "ns3/ipv4-global-routing-helper.h" 

#include <iostream>
#include <fstream>
#include <vector>
#include <string>

using namespace ns3;


void ReceivePacket(Ptr<Socket> socket)
{
    while (socket->Recv())
    {
        NS_LOG_UNCOND("Received one packet!");
    }
}

static void GenerateTraffic(Ptr<Socket> socket, uint32_t pktSize,
        uint32_t pktCount, Time pktInterval)
{
    if (pktCount > 0)
    {
        socket->Send(Create<Packet> (pktSize));
        Simulator::Schedule(pktInterval, &GenerateTraffic,
                socket, pktSize, pktCount - 1, pktInterval);
    }
    else
    {
        socket->Close();
    }
}


int main(int argc, char *argv[])
{
    bool verbose = false;
    std::string g_p2p_lat = "2ms";
    std::string g_p2p_rate = "500kb/s"; // P2P link
    std::string g_wifiPhyMode("DsssRate1Mbps");
    InternetStackHelper g_internetStkHlpr;
    
    uint32_t packetSize = 1000; // bytes
    uint32_t numPackets = 1;
    double interval = 1.0; // seconds
    Time interPacketInterval = Seconds(interval);
    int stationCount=1;
    
    
    
    
    /**----------------------------------------------------------------*/
    
    YansWifiPhyHelper wifiPhyHlpr = YansWifiPhyHelper::Default();
    YansWifiChannelHelper wifiChnlHlpr = YansWifiChannelHelper::Default();
    Ptr<YansWifiChannel> chanel = wifiChnlHlpr.Create();

    wifiPhyHlpr.SetChannel(chanel);

    WifiHelper wifiHlpr;
    if (verbose)
    {
        wifiHlpr.EnableLogComponents(); // Turn on all Wifi logging
    }
    wifiHlpr.SetStandard(WIFI_PHY_STANDARD_80211n_2_4GHZ); // i have changed it to 802.11n2.4GHz

    // Add a non-QoS upper mac, and disable rate control
    NqosWifiMacHelper wifiMacHlpr = NqosWifiMacHelper::Default();
    wifiHlpr.SetRemoteStationManager("ns3::ConstantRateWifiManager",
            "DataMode", StringValue(g_wifiPhyMode),
            "ControlMode", StringValue(g_wifiPhyMode));
    // Set it to adhoc mode
    wifiMacHlpr.SetType("ns3::AdhocWifiMac");
    /**----------------------------------------------------------------*/
    
    
    NodeContainer apNodeContainer;
    apNodeContainer.Create(3);
    NodeContainer nc0_1 = NodeContainer(apNodeContainer.Get(0), apNodeContainer.Get(1));
    NodeContainer nc0_2 = NodeContainer(apNodeContainer.Get(0), apNodeContainer.Get(2));
    
    
    PointToPointHelper p2p;
    p2p.SetDeviceAttribute("DataRate", StringValue(g_p2p_rate));
    p2p.SetChannelAttribute("Delay", StringValue(g_p2p_lat));
    NetDeviceContainer d0_1 = p2p.Install(nc0_1);
    NetDeviceContainer d0_2 = p2p.Install(nc0_2);
    
    
    NodeContainer wifiApNodeContainer;
    wifiApNodeContainer.Add(apNodeContainer.Get(1));
    wifiApNodeContainer.Add(apNodeContainer.Get(2));
    
    NodeContainer wifiStaNodeContainer;
    wifiStaNodeContainer.Create(stationCount);
    
    
    NetDeviceContainer apDevicesContainer = wifiHlpr.Install(wifiPhyHlpr, wifiMacHlpr, wifiApNodeContainer);
    NetDeviceContainer staDevicesContainer = wifiHlpr.Install(wifiPhyHlpr, wifiMacHlpr, wifiStaNodeContainer);
    
    /*Mobility  model start*/
    MobilityHelper mobility;

    mobility.SetPositionAllocator("ns3::GridPositionAllocator",
            "MinX", DoubleValue(0.0),
            "MinY", DoubleValue(0.0),
            "DeltaX", DoubleValue(5.0),
            "DeltaY", DoubleValue(10.0),
            "GridWidth", UintegerValue(3),
            "LayoutType", StringValue("RowFirst"));

    /*
    mobility.SetMobilityModel("ns3::RandomWalk2dMobilityModel",
            "Bounds", RectangleValue(Rectangle(-50, 50, -50, 50)));
    mobility.Install(wifiStaNodes);
     */

    mobility.SetMobilityModel("ns3::ConstantPositionMobilityModel");
    mobility.Install(wifiApNodeContainer);
    mobility.Install(wifiStaNodeContainer);
    /*Mobility model end*/
    
    
    
    
    g_internetStkHlpr.Install(apNodeContainer);
    //g_internetStkHlpr.Install(wifiApNodeContainer);
    g_internetStkHlpr.Install(wifiStaNodeContainer);
    Ipv4AddressHelper ipv4Hlpr;

    ipv4Hlpr.SetBase("10.1.1.0", "255.255.255.0");
    //Ipv4InterfaceContainer p2pInterfaces;
    //p2pInterfaces = address.Assign(d0_1);
    ipv4Hlpr.Assign(d0_1);
    ipv4Hlpr.Assign(d0_2);
    
    ipv4Hlpr.SetBase("10.1.2.0", "255.255.255.0");
    ipv4Hlpr.Assign(apDevicesContainer);
    ipv4Hlpr.Assign(staDevicesContainer);
    
    
    
    TypeId tid = TypeId::LookupByName("ns3::UdpSocketFactory");
    Ptr<Socket> recvSink = Socket::CreateSocket(apNodeContainer.Get(0), tid);
    InetSocketAddress local = InetSocketAddress(Ipv4Address::GetAny(), 80);
    recvSink->Bind(local);
    recvSink->SetRecvCallback(MakeCallback(&ReceivePacket));

    //Ptr<Socket> source = Socket::CreateSocket(wifiStaNodeContainer.Get(0), tid);
    Ptr<Socket> source = Socket::CreateSocket(apNodeContainer.Get(1), tid);
    InetSocketAddress remote = InetSocketAddress(Ipv4Address("255.255.255.255"), 80);
    source->SetAllowBroadcast(true);
    source->Connect(remote);

    //Ipv4GlobalRoutingHelper::PopulateRoutingTables();


    Simulator::ScheduleWithContext(source->GetNode()->GetId(),
            Seconds(1.0), &GenerateTraffic,
            source, packetSize, numPackets, interPacketInterval);

    
    AsciiTraceHelper ascii;
    p2p.EnableAsciiAll(ascii.CreateFileStream("adhoc_test_2.tr"));
    
    
    Simulator::Run();
    Simulator::Destroy();

    return 0;
    
    
}