
#include "ns3/core-module.h"
#include "ns3/network-module.h"
#include "ns3/mobility-module.h"
#include "ns3/config-store-module.h"
#include "ns3/wifi-module.h"
#include "ns3/internet-module.h"
#include <fstream>
#include <string>
#include "ns3/point-to-point-module.h"
#include "ns3/applications-module.h"
#include "ns3/flow-monitor-module.h"
#include "ns3/ipv4-global-routing-helper.h" 

#include <iostream>
#include <fstream>
#include <vector>
#include <string>

using namespace ns3;

void RcvBySink(Ptr<const Packet> pkt, const Address &)
{
    std::cout << Simulator::Now().GetSeconds() << "\t\t\t\tpkt received, pkt Uid = " << pkt->GetUid() <<"\n";
}

void ReceivePacket(Ptr<Socket> socket)
{
    while (socket->Recv())
    {
        NS_LOG_UNCOND("Received one packet!");
    }
}

static void GenerateTraffic(Ptr<Socket> socket, uint32_t pktSize,
        uint32_t pktCount, Time pktInterval)
{
    if (pktCount > 0)
    {
        socket->Send(Create<Packet> (pktSize));
        Simulator::Schedule(pktInterval, &GenerateTraffic,
                socket, pktSize, pktCount - 1, pktInterval);
    }
    else
    {
        socket->Close();
    }
}

class MyApp : public Application
{
public:

    MyApp();
    virtual ~MyApp();

    void Setup(Ptr<Socket> socket, Address address, uint32_t packetSize, uint32_t nPackets, DataRate dataRate);
    void ChangeRate(DataRate newrate);

private:
    virtual void StartApplication(void);
    virtual void StopApplication(void);

    void ScheduleTx(void);
    void SendPacket(void);

    Ptr<Socket> m_socket;
    Address m_peer;
    uint32_t m_packetSize;
    uint32_t m_nPackets;
    DataRate m_dataRate;
    EventId m_sendEvent;
    bool m_running;
    uint32_t m_packetsSent;
};

MyApp::MyApp()
: m_socket(0),
m_peer(),
m_packetSize(0),
m_nPackets(0),
m_dataRate(0),
m_sendEvent(),
m_running(false),
m_packetsSent(0)
{
}

MyApp::~MyApp()
{
    m_socket = 0;
}

void
MyApp::Setup(Ptr<Socket> socket, Address address, uint32_t packetSize, uint32_t nPackets, DataRate dataRate)
{
    m_socket = socket;
    m_peer = address;
    m_packetSize = packetSize;
    m_nPackets = nPackets;
    m_dataRate = dataRate;
}

void
MyApp::StartApplication(void)
{
    m_running = true;
    m_packetsSent = 0;
    m_socket->Bind();
    m_socket->Connect(m_peer);
    SendPacket();
}

void
MyApp::StopApplication(void)
{
    m_running = false;

    if (m_sendEvent.IsRunning())
    {
        Simulator::Cancel(m_sendEvent);
    }

    if (m_socket)
    {
        m_socket->Close();
    }
}

void
MyApp::SendPacket(void)
{
    Ptr<Packet> packet = Create<Packet> (m_packetSize);
    m_socket->Send(packet);

    if (++m_packetsSent < m_nPackets)
    {
        ScheduleTx();
    }
}

void
MyApp::ScheduleTx(void)
{
    if (m_running)
    {
        Time tNext(Seconds(m_packetSize * 8 / static_cast<double> (m_dataRate.GetBitRate())));
        m_sendEvent = Simulator::Schedule(tNext, &MyApp::SendPacket, this);
    }
}

void
MyApp::ChangeRate(DataRate newrate)
{
    m_dataRate = newrate;
    return;
}

static void
CwndChange(uint32_t oldCwnd, uint32_t newCwnd)
{
    std::cout << Simulator::Now().GetSeconds() << "\t" << newCwnd << "\n";
}

void
IncRate(Ptr<MyApp> app, DataRate rate)
{
    app->ChangeRate(rate);
    return;
}

int main(int argc, char *argv[])
{
    bool verbose = false;
    std::string g_p2p_lat = "2ms";
    std::string g_p2p_rate = "500kb/s"; // P2P link
    std::string g_wifiPhyMode("DsssRate1Mbps");
    InternetStackHelper g_internetStkHlpr;

    int stationCount = 1;




    /**----------------------------------------------------------------*/

    YansWifiPhyHelper wifiPhyHlpr = YansWifiPhyHelper::Default();
    YansWifiChannelHelper wifiChnlHlpr = YansWifiChannelHelper::Default();
    Ptr<YansWifiChannel> chanel = wifiChnlHlpr.Create();

    wifiPhyHlpr.SetChannel(chanel);

    WifiHelper wifiHlpr;
    if (verbose)
    {
        wifiHlpr.EnableLogComponents(); // Turn on all Wifi logging
    }
    wifiHlpr.SetStandard(WIFI_PHY_STANDARD_80211n_2_4GHZ); // i have changed it to 802.11n2.4GHz

    // Add a non-QoS upper mac, and disable rate control
    NqosWifiMacHelper wifiMacHlpr = NqosWifiMacHelper::Default();
    wifiHlpr.SetRemoteStationManager("ns3::ConstantRateWifiManager",
            "DataMode", StringValue(g_wifiPhyMode),
            "ControlMode", StringValue(g_wifiPhyMode));
    // Set it to adhoc mode
    wifiMacHlpr.SetType("ns3::AdhocWifiMac");
    /**----------------------------------------------------------------*/




    /*
     * 
    NodeContainer apNodeContainer;
    apNodeContainer.Create(3);
    
    NodeContainer nc0_1 = NodeContainer(apNodeContainer.Get(0), apNodeContainer.Get(1));
    NodeContainer nc0_2 = NodeContainer(apNodeContainer.Get(0), apNodeContainer.Get(2));
    
    
    PointToPointHelper p2p;
    p2p.SetDeviceAttribute("DataRate", StringValue(g_p2p_rate));
    p2p.SetChannelAttribute("Delay", StringValue(g_p2p_lat));
    NetDeviceContainer d0_1 = p2p.Install(nc0_1);
    NetDeviceContainer d0_2 = p2p.Install(nc0_2);
     */

    NodeContainer wifiApNodeContainer;
    wifiApNodeContainer.Create(1);

    NodeContainer wifiStaNodeContainer;
    wifiStaNodeContainer.Create(stationCount);


    NetDeviceContainer apDevicesContainer = wifiHlpr.Install(wifiPhyHlpr, wifiMacHlpr, wifiApNodeContainer);
    NetDeviceContainer staDevicesContainer = wifiHlpr.Install(wifiPhyHlpr, wifiMacHlpr, wifiStaNodeContainer);

    /*Mobility  model start*/
    MobilityHelper mobility;

    mobility.SetPositionAllocator("ns3::GridPositionAllocator",
            "MinX", DoubleValue(0.0),
            "MinY", DoubleValue(0.0),
            "DeltaX", DoubleValue(5.0),
            "DeltaY", DoubleValue(10.0),
            "GridWidth", UintegerValue(3),
            "LayoutType", StringValue("RowFirst"));

    /*
    mobility.SetMobilityModel("ns3::RandomWalk2dMobilityModel",
            "Bounds", RectangleValue(Rectangle(-50, 50, -50, 50)));
    mobility.Install(wifiStaNodes);
     */

    mobility.SetMobilityModel("ns3::ConstantPositionMobilityModel");
    mobility.Install(wifiApNodeContainer);
    mobility.Install(wifiStaNodeContainer);
    /*Mobility model end*/




    //g_internetStkHlpr.Install(apNodeContainer);
    g_internetStkHlpr.Install(wifiApNodeContainer);
    g_internetStkHlpr.Install(wifiStaNodeContainer);
    Ipv4AddressHelper ipv4Hlpr;

    /*    
        ipv4Hlpr.SetBase("10.1.1.0", "255.255.255.0");
        Ipv4InterfaceContainer d0_1_InterfCntnr = ipv4Hlpr.Assign(d0_1);
        Ipv4InterfaceContainer d0_2_InterfCntnr = ipv4Hlpr.Assign(d0_2);
     */

    ipv4Hlpr.SetBase("10.1.2.0", "255.255.255.0");
    Ipv4InterfaceContainer apInterfaceContainer = ipv4Hlpr.Assign(apDevicesContainer);
    Ipv4InterfaceContainer staInterfaceContainer = ipv4Hlpr.Assign(staDevicesContainer);

#if 1
    //NS_LOG_INFO("Create Applications.");

    // TCP connfection from N0 to N2

    int sinkNodeIncex = 0;
    uint16_t sinkPort = 8080;
    Address sinkAddress(InetSocketAddress(apInterfaceContainer.GetAddress(sinkNodeIncex), sinkPort)); // interface of n2
    PacketSinkHelper packetSinkHelper("ns3::TcpSocketFactory", InetSocketAddress(Ipv4Address::GetAny(), sinkPort));
    ApplicationContainer sinkApps = packetSinkHelper.Install(wifiApNodeContainer.Get(sinkNodeIncex)); //n2 as sink
    sinkApps.Start(Seconds(0.));
    sinkApps.Stop(Seconds(100.));

    int sourceNodeIncex = 0;
    Ptr<Socket> ns3TcpSocket = Socket::CreateSocket(wifiStaNodeContainer.Get(sourceNodeIncex), TcpSocketFactory::GetTypeId()); //source at n0

    // Trace Congestion window
    ns3TcpSocket->TraceConnectWithoutContext("CongestionWindow", MakeCallback(&CwndChange));
    //std::ostringstream oss;

    Config::ConnectWithoutContext("/NodeList/*/ApplicationList/*/$ns3::PacketSink/Rx", MakeCallback(&RcvBySink));

    // Create TCP application at n0
    Ptr<MyApp> app = CreateObject<MyApp> ();
    app->Setup(ns3TcpSocket, sinkAddress, 1040, 100000, DataRate("250Kbps"));
    wifiStaNodeContainer.Get(sourceNodeIncex)->AddApplication(app);
    app->SetStartTime(Seconds(1.));
    app->SetStopTime(Seconds(100.));

#else 
    uint32_t packetSize = 1000; // bytes
    uint32_t numPackets = 1;
    double interval = 1.0; // seconds
    Time interPacketInterval = Seconds(interval);


    TypeId tid = TypeId::LookupByName("ns3::UdpSocketFactory");
    Ptr<Socket> recvSink = Socket::CreateSocket(wifiApNodeContainer.Get(0), tid);
    InetSocketAddress local = InetSocketAddress(Ipv4Address::GetAny(), 80);
    recvSink->Bind(local);
    recvSink->SetRecvCallback(MakeCallback(&ReceivePacket));

    //Ptr<Socket> source = Socket::CreateSocket(wifiStaNodeContainer.Get(0), tid);
    Ptr<Socket> source = Socket::CreateSocket(wifiStaNodeContainer.Get(0), tid);
    InetSocketAddress remote = InetSocketAddress(Ipv4Address("255.255.255.255"), 80);
    source->SetAllowBroadcast(true);
    source->Connect(remote);

    //Ipv4GlobalRoutingHelper::PopulateRoutingTables();

    Simulator::ScheduleWithContext(source->GetNode()->GetId(),
            Seconds(1.0), &GenerateTraffic,
            source, packetSize, numPackets, interPacketInterval);
#endif 


    wifiPhyHlpr.EnablePcap("pureAdhoc", staDevicesContainer.Get(0), true);



    Simulator::Stop(Seconds(10.0));
    Simulator::Run();
    Simulator::Destroy();

    return 0;


}