
#include "ns3/core-module.h"
#include "ns3/network-module.h"
#include "ns3/mobility-module.h"
#include "ns3/config-store-module.h"
#include "ns3/wifi-module.h"
#include "ns3/internet-module.h"
#include <fstream>
#include <string>
#include "ns3/point-to-point-module.h"
#include "ns3/applications-module.h"
#include "ns3/flow-monitor-module.h"
#include "ns3/ipv4-global-routing-helper.h" 

#include <iostream>
#include <fstream>
#include <vector>
#include <string>

using namespace ns3;

void RcvBySink(Ptr<const Packet> pkt, const Address &)
{
    //std::cout << Simulator::Now().GetSeconds() << "\t\t\t\tpkt received, pkt Uid = " << pkt->GetUid() << "\n";
}

void ReceivePacket(Ptr<Socket> socket)
{
    while (socket->Recv())
    {
        NS_LOG_UNCOND("Received one packet!");
    }
}

static void GenerateTraffic(Ptr<Socket> socket, uint32_t pktSize,
        uint32_t pktCount, Time pktInterval)
{
    if (pktCount > 0)
    {
        socket->Send(Create<Packet> (pktSize));
        Simulator::Schedule(pktInterval, &GenerateTraffic,
                socket, pktSize, pktCount - 1, pktInterval);
    }
    else
    {
        socket->Close();
    }
}

class MyApp : public Application
{
public:

    MyApp();
    virtual ~MyApp();

    void Setup(Ptr<Socket> socket, Address address, uint32_t packetSize, uint32_t nPackets, DataRate dataRate);
    void ChangeRate(DataRate newrate);

private:
    virtual void StartApplication(void);
    virtual void StopApplication(void);

    void ScheduleTx(void);
    void SendPacket(void);

    Ptr<Socket> m_socket;
    Address m_peer;
    uint32_t m_packetSize;
    uint32_t m_nPackets;
    DataRate m_dataRate;
    EventId m_sendEvent;
    bool m_running;
    uint32_t m_packetsSent;
};

MyApp::MyApp()
: m_socket(0),
m_peer(),
m_packetSize(0),
m_nPackets(0),
m_dataRate(0),
m_sendEvent(),
m_running(false),
m_packetsSent(0)
{
}

MyApp::~MyApp()
{
    m_socket = 0;
}

void
MyApp::Setup(Ptr<Socket> socket, Address address, uint32_t packetSize, uint32_t nPackets, DataRate dataRate)
{
    m_socket = socket;
    m_peer = address;
    m_packetSize = packetSize;
    m_nPackets = nPackets;
    m_dataRate = dataRate;
}

void
MyApp::StartApplication(void)
{
    m_running = true;
    m_packetsSent = 0;
    m_socket->Bind();
    m_socket->Connect(m_peer);
    SendPacket();
}

void
MyApp::StopApplication(void)
{
    m_running = false;

    if (m_sendEvent.IsRunning())
    {
        Simulator::Cancel(m_sendEvent);
    }

    if (m_socket)
    {
        m_socket->Close();
    }
}

void
MyApp::SendPacket(void)
{
    Ptr<Packet> packet = Create<Packet> (m_packetSize);
    m_socket->Send(packet);

    if (++m_packetsSent < m_nPackets)
    {
        ScheduleTx();
    }
}

void
MyApp::ScheduleTx(void)
{
    if (m_running)
    {
        Time tNext(Seconds(m_packetSize * 8 / static_cast<double> (m_dataRate.GetBitRate())));
        m_sendEvent = Simulator::Schedule(tNext, &MyApp::SendPacket, this);
    }
}

void
MyApp::ChangeRate(DataRate newrate)
{
    m_dataRate = newrate;
    return;
}

/*
static void
CwndChange(uint32_t oldCwnd, uint32_t newCwnd)
{
    std::cout << Simulator::Now().GetSeconds() << "\t" << newCwnd << "\n";
}
*/
void
IncRate(Ptr<MyApp> app, DataRate rate)
{
    app->ChangeRate(rate);
    return;
}

static void PhyTxBegin(Ptr<const Packet> p)
{
    //NS_LOG_UNCOND("\t@PhyTxBegin at "<< Simulator::Now().GetSeconds());
}
static void PhyTxEnd(Ptr<const Packet> p)
{
    //NS_LOG_UNCOND("\t@PhyTxEnd at "<< Simulator::Now().GetSeconds());
}

static void PhyRxBegin(Ptr<const Packet> p)
{
    //NS_LOG_UNCOND("\t\t@PhyRxBegin at "<< Simulator::Now().GetSeconds());
}

static void PhyRxEnd(Ptr<const Packet> p)
{
    //NS_LOG_UNCOND("\t\t@PhyRxEnd at "<< Simulator::Now().GetSeconds());
}

static void PhyTxDrop(Ptr<const Packet> p)
{
    //NS_LOG_UNCOND("\t\t\t@PhyTxDrop at "<< Simulator::Now().GetSeconds());
}

static void PhyRxDrop(Ptr<const Packet> p)
{
    //NS_LOG_UNCOND("\t\t\t@PhyRxDrop at "<< Simulator::Now().GetSeconds());
}

int main(int argc, char *argv[])
{
    bool verbose = false;
    
    CommandLine cmd;
    cmd.Parse(argc, argv);

    NodeContainer wifiNodeContainer;
    wifiNodeContainer.Create(2);
    
    WifiHelper wifiHlpr;
    if (verbose)
    {
        wifiHlpr.EnableLogComponents(); // Turn on all Wifi logging
    }
    //wifiHlpr.SetStandard(WIFI_PHY_STANDARD_80211n_2_4GHZ); // i have changed it to 802.11n2.4GHz
    wifiHlpr.SetStandard(WIFI_PHY_STANDARD_80211g); // i have changed it to 802.11n2.4GHz
    //wifiHlpr.SetStandard(WIFI_PHY_STANDARD_80211a); // i have changed it to 802.11n2.4GHz
    
    
    
    // Add a non-QoS upper mac, and disable rate control
    
    // Set to a non-QoS upper mac
    NqosWifiMacHelper wifiMacHlpr = NqosWifiMacHelper::Default();
    
    // Set it to adhoc mode
    wifiMacHlpr.SetType("ns3::AdhocWifiMac");
    
    //std::string g_wifiPhyMode("DsssRate1Mbps");
    //std::string g_wifiPhyMode("ErpOfdmRate54Mbps");
    std::string g_wifiPhyControlMode("ErpOfdmRate54Mbps");
    std::string g_wifiPhyDataMode("ErpOfdmRate54Mbps");
    //std::string g_wifiPhyControlMode("OfdmRate54Mbps");
    //std::string g_wifiPhyDataMode("OfdmRate54Mbps");
    // Set Wi-Fi rate manager
    wifiHlpr.SetRemoteStationManager("ns3::ConstantRateWifiManager",
            "DataMode", StringValue(g_wifiPhyDataMode),
            "ControlMode", StringValue(g_wifiPhyControlMode));
    
    double simulationTimeSec = 3;


    // disable rts cts all the time.
    Config::SetDefault("ns3::WifiRemoteStationManager::RtsCtsThreshold", StringValue("2200"));
    // disable fragmentation
    Config::SetDefault("ns3::WifiRemoteStationManager::FragmentationThreshold", StringValue("2200"));


    /**----------------------------------------------------------------*/

    YansWifiPhyHelper wifiPhyHlpr = YansWifiPhyHelper::Default();
    YansWifiChannelHelper wifiChnlHlpr = YansWifiChannelHelper::Default();
    Ptr<YansWifiChannel> chanel = wifiChnlHlpr.Create();

    wifiPhyHlpr.SetChannel(chanel);
    
    /**----------------------------------------------------------------*/



    
    NetDeviceContainer wifiDevicesContainer = wifiHlpr.Install(wifiPhyHlpr, wifiMacHlpr, wifiNodeContainer);

    /*Mobility  model start*/
    MobilityHelper mobility;

    mobility.SetPositionAllocator("ns3::GridPositionAllocator",
            "MinX", DoubleValue(0.0),
            "MinY", DoubleValue(0.0),
            "DeltaX", DoubleValue(10.0),
            "DeltaY", DoubleValue(10.0),
            "GridWidth", UintegerValue(3),
            "LayoutType", StringValue("RowFirst"));

    /*
    mobility.SetMobilityModel("ns3::RandomWalk2dMobilityModel",
            "Bounds", RectangleValue(Rectangle(-50, 50, -50, 50)));
    mobility.Install(wifiStaNodes);
     */

    mobility.SetMobilityModel("ns3::ConstantPositionMobilityModel");
    mobility.Install(wifiNodeContainer);
    /*Mobility model end*/

    InternetStackHelper g_internetStkHlpr;
    g_internetStkHlpr.Install(wifiNodeContainer);


    Ipv4AddressHelper ipv4Hlpr;

    ipv4Hlpr.SetBase("10.1.1.0", "255.255.255.0");
    Ipv4InterfaceContainer wifiInterfaceContainer = ipv4Hlpr.Assign(wifiDevicesContainer);


    NS_LOG_UNCOND("Create Applications.");



    uint32_t UDPpacketSize = 1500; // bytes
    uint32_t UDPnumPackets = 9999999;
    int senderID = 0;
    int receiverID = 1;

    uint16_t sinkPort = 1010;
    Address staSinkAddress;
    staSinkAddress = InetSocketAddress(wifiInterfaceContainer.GetAddress(receiverID), sinkPort); // interface of sink
    PacketSinkHelper packetSinkHelper("ns3::UdpSocketFactory", InetSocketAddress(Ipv4Address::GetAny(), sinkPort));
    ApplicationContainer sinkApps = packetSinkHelper.Install(wifiNodeContainer.Get(receiverID));
    sinkApps.Start(Seconds(0.0));
    sinkApps.Stop(Seconds(simulationTimeSec));

    Ptr<Socket> ns3UdpSocket = Socket::CreateSocket(wifiNodeContainer.Get(senderID), UdpSocketFactory::GetTypeId());

    Ptr<MyApp> app = CreateObject<MyApp> ();
    app->Setup(ns3UdpSocket, staSinkAddress, UDPpacketSize, UDPnumPackets, DataRate("54Mbps"));
    wifiNodeContainer.Get(senderID)->AddApplication(app);
    //Config::ConnectWithoutContext("/NodeList/*/ApplicationList/*/$ns3::PacketSink/Rx", MakeCallback(&RcvBySink));
    app->SetStartTime(Seconds(0.0));
    app->SetStopTime(Seconds(simulationTimeSec));


    Ipv4GlobalRoutingHelper::PopulateRoutingTables();

    wifiPhyHlpr.EnablePcap("Adhoc_41_wifiOnly_", wifiDevicesContainer.Get(0), true);

    Config::ConnectWithoutContext("/NodeList/*/ApplicationList/*/$ns3::PacketSink/Rx", MakeCallback(&RcvBySink));
    Config::ConnectWithoutContext("/NodeList/*/DeviceList/*/$ns3::WifiNetDevice/Phy/PhyTxBegin", MakeCallback(&PhyTxBegin));
    Config::ConnectWithoutContext("/NodeList/*/DeviceList/*/$ns3::WifiNetDevice/Phy/PhyTxEnd", MakeCallback(&PhyTxEnd));
    Config::ConnectWithoutContext("/NodeList/*/DeviceList/*/$ns3::WifiNetDevice/Phy/PhyRxBegin", MakeCallback(&PhyRxBegin));
    Config::ConnectWithoutContext("/NodeList/*/DeviceList/*/$ns3::WifiNetDevice/Phy/PhyRxEnd", MakeCallback(&PhyRxEnd));
    Config::ConnectWithoutContext("/NodeList/*/DeviceList/*/$ns3::WifiNetDevice/Phy/PhyTxDrop", MakeCallback(&PhyTxDrop));
    Config::ConnectWithoutContext("/NodeList/*/DeviceList/*/$ns3::WifiNetDevice/Phy/PhyRxDrop", MakeCallback(&PhyRxDrop));

    Simulator::setTestVariable(123);
    NS_LOG_UNCOND(Simulator::getTestVariable());

    Simulator::Stop(Seconds(simulationTimeSec));
    Simulator::Run();
    Simulator::Destroy();
    
  
    uint32_t totalUDPpacketsThrough = DynamicCast<PacketSink> (sinkApps.Get (0))->GetTotalRx ();
    double UDPthroughput = totalUDPpacketsThrough * 8 / (simulationTimeSec * 1000000.0); //Mbit/s
    //std::cout << totalUDPpacketsThrough << " Mbps\n";
    std::cout <<"total UDP byte : " << totalUDPpacketsThrough << "; UDP throughput in AP :  " << UDPthroughput << " Mbps\n";

    return 0;


}
