//https://www.nsnam.org/doxygen/ht-wifi-network_8cc_source.html 

#include "ns3/core-module.h"
#include "ns3/network-module.h"
#include "ns3/mobility-module.h"
#include "ns3/config-store-module.h"
#include "ns3/wifi-module.h"
#include "ns3/csma-module.h"
#include "ns3/internet-module.h"
#include <fstream>
#include <string>
#include "ns3/point-to-point-module.h"
#include "ns3/applications-module.h"
#include "ns3/flow-monitor-module.h"
#include "ns3/ipv4-global-routing-helper.h"
#include "src/core/model/simulator.h"
#include "src/network/model/node.h"
#include "src/network/model/net-device.h"
#include "src/network/model/packet.h" 

#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <stdio.h>

using namespace ns3;

///////////////////////////////////////////////////////////////////////////////////////////////
/*
void pktNodeidDropCallback(Ptr<Packet> pkt, uint32_t nodeID)
{
    NS_LOG_UNCOND("pkt id "<< pkt->GetUid() << ", Dropped at nodeID : "<< nodeID);
}
 */
///////////////////////////////////////////////////////////////////////////////////////////////
static void PhyTxBegin(Ptr<const Packet> p)
{
    NS_LOG_UNCOND("\t@PhyTxBegin at "<< Simulator::Now().GetSeconds());
}
static void PhyTxEnd(Ptr<const Packet> p)
{
    NS_LOG_UNCOND("\t@PhyTxEnd at "<< Simulator::Now().GetSeconds());
}

static void PhyRxBegin(Ptr<const Packet> p)
{
    NS_LOG_UNCOND("\t\t@PhyRxBegin at "<< Simulator::Now().GetSeconds());
}

static void PhyRxEnd(Ptr<const Packet> p)
{
    NS_LOG_UNCOND("\t\t@PhyRxEnd at "<< Simulator::Now().GetSeconds());
}

static void PhyTxDrop(Ptr<const Packet> p)
{
    NS_LOG_UNCOND("\t\t\t@PhyTxDrop at "<< Simulator::Now().GetSeconds());
}

static void PhyRxDrop(Ptr<const Packet> p)
{
    NS_LOG_UNCOND("\t\t\t@PhyRxDrop at "<< Simulator::Now().GetSeconds());
}
/////////////////////////////////////////////////////////////////////////////////////////////////
static void RcvBySink(Ptr<const Packet> pkt, const Address &)
{
    //std::cout << Simulator::Now().GetSeconds() << "\t\t\t\tpkt received, pkt Uid = " << pkt->GetUid() << "\n";
}

void ReceivePacket(Ptr<Socket> socket)
{
    while (socket->Recv())
    {
        NS_LOG_UNCOND("Received one packet!");
    }
}

static void GenerateTraffic(Ptr<Socket> socket, uint32_t pktSize,
        uint32_t pktCount, Time pktInterval)
{
    if (pktCount > 0)
    {
        socket->Send(Create<Packet> (pktSize));
        Simulator::Schedule(pktInterval, &GenerateTraffic,
                socket, pktSize, pktCount - 1, pktInterval);
    }
    else
    {
        socket->Close();
    }
}

class MyApp : public Application
{
public:

    MyApp();
    virtual ~MyApp();

    void Setup(Ptr<Socket> socket, Address address, uint32_t packetSize, uint32_t nPackets, DataRate dataRate);
    void ChangeRate(DataRate newrate);

private:
    virtual void StartApplication(void);
    virtual void StopApplication(void);

    void ScheduleTx(void);
    void SendPacket(void);

    Ptr<Socket> m_socket;
    Address m_peer;
    uint32_t m_packetSize;
    uint32_t m_nPackets;
    DataRate m_dataRate;
    EventId m_sendEvent;
    bool m_running;
    uint32_t m_packetsSent;
};

MyApp::MyApp()
: m_socket(0),
m_peer(),
m_packetSize(0),
m_nPackets(0),
m_dataRate(0),
m_sendEvent(),
m_running(false),
m_packetsSent(0)
{
}

MyApp::~MyApp()
{
    m_socket = 0;
}

void
MyApp::Setup(Ptr<Socket> socket, Address address, uint32_t packetSize, uint32_t nPackets, DataRate dataRate)
{
    m_socket = socket;
    m_peer = address;
    m_packetSize = packetSize;
    m_nPackets = nPackets;
    m_dataRate = dataRate;
}

void
MyApp::StartApplication(void)
{
    m_running = true;
    m_packetsSent = 0;
    m_socket->Bind();
    m_socket->Connect(m_peer);
    SendPacket();
}

void
MyApp::StopApplication(void)
{
    m_running = false;

    if (m_sendEvent.IsRunning())
    {
        Simulator::Cancel(m_sendEvent);
    }

    if (m_socket)
    {
        m_socket->Close();
    }
}

void
MyApp::SendPacket(void)
{
    Ptr<Packet> packet = Create<Packet> (m_packetSize);
    m_socket->Send(packet);

    if (++m_packetsSent < m_nPackets)
    {
        ScheduleTx();
    }
}

void
MyApp::ScheduleTx(void)
{
    if (m_running)
    {
        Time tNext(Seconds(m_packetSize * 8 / static_cast<double> (m_dataRate.GetBitRate())));
        m_sendEvent = Simulator::Schedule(tNext, &MyApp::SendPacket, this);
    }
}

void
MyApp::ChangeRate(DataRate newrate)
{
    m_dataRate = newrate;
    return;
}

void
IncRate(Ptr<MyApp> app, DataRate rate)
{
    app->ChangeRate(rate);
    return;
}

static void
CwndChange(uint32_t oldCwnd, uint32_t newCwnd)
{
    //std::cout << Simulator::Now().GetSeconds() << "\t" << newCwnd << "\n";
}

enum SINK_TYPE_Plaban
{
    TCP_TYPE = 10,
    UDP_TYPE = 11,
};

class AP_STA_Cluster : public Object
{
public:
    double simulationTimeSec;
    int m_wifi_stationCount;
    int m_numberOfAPchannel;
    int m_numberOfNodeINcsmaAPcluster;
    int groupID;
    int ap_x;
    int ap_y;
    int ap_txRadious;
    CHANNEL_Plaban defaultChannel;
    Ptr<UniformRandomVariable> rand;

    CsmaHelper csmaHlpr;

    NodeContainer apCsmaNodeContainer;
    NodeContainer wifiApNodeContainer;
    NodeContainer wifiStaNodeContainer;

    NetDeviceContainer apDevicesContainer;
    NetDeviceContainer staDevicesContainer;
    NetDeviceContainer csmaNetDeviceContainer;

    Ipv4InterfaceContainer csmaIpv4IntfcCntnr;
    Ipv4InterfaceContainer apInterfaceContainer;
    Ipv4InterfaceContainer staInterfaceContainer;

    //ApplicationContainer apTCPsinkApps;
    //ApplicationContainer apUDPsinkApps;

    Address apTCPsinkAddress;
    Address apUDPsinkAddress;

    std::map< uint32_t, std::map<SINK_TYPE_Plaban, ApplicationContainer> > nodeID_sinkType_AppContainer_MAP;

    AP_STA_Cluster()
    {
        this->m_numberOfAPchannel = 3; // WARNIGN : if you change it, then change "enum CHANNEL_Plaban" accordingly... Also dont forget to assign individual channel for this->wifiApNodeContainer
        this->m_numberOfNodeINcsmaAPcluster = 1 + this->m_numberOfAPchannel;
    }

    void insert_nodeID_SinkType_AppContainer(uint32_t _nodeID, SINK_TYPE_Plaban _sinkType, ApplicationContainer & _appContainer)
    {
        std::map< uint32_t, std::map<SINK_TYPE_Plaban, ApplicationContainer> >::iterator itNode;

        itNode = this->nodeID_sinkType_AppContainer_MAP.find(_nodeID);

        if (this->nodeID_sinkType_AppContainer_MAP.end() == itNode)
        {
            std::map<SINK_TYPE_Plaban, ApplicationContainer> sinktype_appContainer_map;
            sinktype_appContainer_map.insert(std::pair<SINK_TYPE_Plaban, ApplicationContainer>(_sinkType, _appContainer));

            this->nodeID_sinkType_AppContainer_MAP.insert(std::pair< uint32_t, std::map<SINK_TYPE_Plaban, ApplicationContainer> >(_nodeID, sinktype_appContainer_map));
        }
        else
        {
            std::pair < std::map<SINK_TYPE_Plaban, ApplicationContainer>::iterator, bool > ret;
            ret = itNode->second.insert(std::pair<SINK_TYPE_Plaban, ApplicationContainer>(_sinkType, _appContainer));

            if (false == ret.second)
            {// already exists
                NS_ASSERT(false);
            }
        }
    }

    ApplicationContainer getAppContainer_givenNodeID_SinkType(uint32_t _nodeID, SINK_TYPE_Plaban _sinkType)
    {
        std::map< uint32_t, std::map<SINK_TYPE_Plaban, ApplicationContainer> >::iterator itNode;

        itNode = this->nodeID_sinkType_AppContainer_MAP.find(_nodeID);

        NS_ASSERT(this->nodeID_sinkType_AppContainer_MAP.end() != itNode); // error if not found

        std::map<SINK_TYPE_Plaban, ApplicationContainer>::iterator it2;
        it2 = itNode->second.find(_sinkType);

        NS_ASSERT(itNode->second.end() != it2); // error if not found

        return it2->second;
    }

    void throughputInAllSinkSocket()
    {
        std::map< uint32_t, std::map<SINK_TYPE_Plaban, ApplicationContainer> >::iterator itNode;

        for (itNode = this->nodeID_sinkType_AppContainer_MAP.begin(); this->nodeID_sinkType_AppContainer_MAP.end() != itNode; ++itNode)
        {
            uint32_t nodeID = itNode->first;

            std::map<SINK_TYPE_Plaban, ApplicationContainer>::iterator itSinkType;
            NS_LOG_UNCOND("");
            for (itSinkType = itNode->second.begin(); itNode->second.end() != itSinkType; ++itSinkType)
            {
                uint32_t totalReceivedByte = DynamicCast<PacketSink> (itSinkType->second.Get(0))->GetTotalRx();
                double throughput_Mbps = (8 * totalReceivedByte) / (this->simulationTimeSec * 1000000.0);

                switch (itSinkType->first)
                {
                    case TCP_TYPE:
                    {
                        NS_LOG_UNCOND("Node " << nodeID << ", TCP sink throughput " << throughput_Mbps << " Mbps");
                        break;
                    }
                    case UDP_TYPE:
                    {
                        NS_LOG_UNCOND("Node " << nodeID << ", UDP sink throughput " << throughput_Mbps << " Mbps");
                        break;
                    }
                    default:
                    {
                        NS_ASSERT(false);
                    }
                }


            }

        }

    }

    void sba_print_Chnl_Mac_Relation(std::map< int, std::map<CHANNEL_Plaban, Mac48Address> > sba_wifiAP_GroupBasedMacChnlMap, std::map< int, std::map<CHANNEL_Plaban, Mac48Address> > sba_csmaGroupBasedMacChnlMap)
    {
        std::map< int, std::map<CHANNEL_Plaban, Mac48Address> >::iterator it_WIFI;
        std::map< int, std::map<CHANNEL_Plaban, Mac48Address> >::iterator it_CSMA;

        NS_LOG_UNCOND("\n++++++++++++++++++ wifi MAC-CHNL +++++++++++++++++++");
        for (it_WIFI = sba_wifiAP_GroupBasedMacChnlMap.begin(); sba_wifiAP_GroupBasedMacChnlMap.end() != it_WIFI; ++it_WIFI)
        {

            NS_LOG_UNCOND("----------------------- GROUP ID " << it_WIFI->first << " ----------------------------");
            std::map<CHANNEL_Plaban, Mac48Address>::iterator itChnlMac;
            for (itChnlMac = it_WIFI->second.begin(); it_WIFI->second.end() != itChnlMac; ++itChnlMac)
            {
                NS_LOG_UNCOND("Channel " << itChnlMac->first << " ; MAC " << itChnlMac->second);
            }
        }
        NS_LOG_UNCOND("\n----------------------------------------------------");


        NS_LOG_UNCOND("\n++++++++++++++++++ CSMA MAC-CHNL +++++++++++++++++++");
        for (it_CSMA = sba_csmaGroupBasedMacChnlMap.begin(); sba_csmaGroupBasedMacChnlMap.end() != it_CSMA; ++it_CSMA)
        {
            NS_LOG_UNCOND("----------------------- GROUP ID " << it_CSMA->first << " ----------------------------");
            std::map<CHANNEL_Plaban, Mac48Address>::iterator itChnlMac;
            for (itChnlMac = it_CSMA->second.begin(); it_CSMA->second.end() != itChnlMac; ++itChnlMac)
            {
                NS_LOG_UNCOND("Channel " << itChnlMac->first << " ; MAC " << itChnlMac->second);
            }
        }
        NS_LOG_UNCOND("\n----------------------------------------------------\n");
    }
    
    void sba_print_Chnl_IPAdr_Relation(std::map< int, std::map<CHANNEL_Plaban, Ipv4Address> > sba_wifiGroupBased_IPAdr_ChnlMap, std::map< int, std::map<CHANNEL_Plaban, Ipv4Address> > sba_csmaGroupBased_IPAdr_ChnlMap)
    {
        std::map< int, std::map<CHANNEL_Plaban, Ipv4Address> >::iterator it;
        NS_LOG_UNCOND("\n++++++++++++++++++ wifi IPAdr-CHNL +++++++++++++++++++");
        for(it = sba_wifiGroupBased_IPAdr_ChnlMap.begin() ; sba_wifiGroupBased_IPAdr_ChnlMap.end() != it ; ++it)
        {
            NS_LOG_UNCOND("----------------------- GROUP ID " << it->first << " ----------------------------");
            std::map<CHANNEL_Plaban, Ipv4Address>::iterator it2;
            for(it2 = it->second.begin() ; it->second.end() != it2 ; ++it2)
            {
                NS_LOG_UNCOND("Channel " << it2->first << " ; IP " << it2->second);
            }
        }
        NS_LOG_UNCOND("\n----------------------------------------------------");
        
        NS_LOG_UNCOND("\n++++++++++++++++++ CSMA IPAdr-CHNL +++++++++++++++++++");
        for(it = sba_csmaGroupBased_IPAdr_ChnlMap.begin() ; sba_csmaGroupBased_IPAdr_ChnlMap.end() != it ; ++it)
        {
            NS_LOG_UNCOND("----------------------- GROUP ID " << it->first << " ----------------------------");
            std::map<CHANNEL_Plaban, Ipv4Address>::iterator it2;
            for(it2 = it->second.begin() ; it->second.end() != it2 ; ++it2)
            {
                NS_LOG_UNCOND("Channel " << it2->first << " ; IP " << it2->second);
            }
        }
        NS_LOG_UNCOND("\n----------------------------------------------------");
    }

    void setup(int _groupID, CHANNEL_Plaban _defaultChannel, double _simulationTimeSec,
    int _m_wifi_stationCount, WifiHelper & wifiHlpr, YansWifiPhyHelper & wifiPhyHlpr,
            NqosWifiMacHelper & wifiMacHlpr, Ipv4AddressHelper & ipv4Hlpr,
            InternetStackHelper & g_internetStkHlpr, uint32_t APx, uint32_t APy, uint32_t APtxRadious, Ptr<UniformRandomVariable> _rand)
    {
        this->groupID = _groupID;
        this->defaultChannel = _defaultChannel;

        this->ap_x = APx;
        this->ap_y = APy;
        this->ap_txRadious = APtxRadious;
        this->rand = _rand;
        
        this->simulationTimeSec = _simulationTimeSec;
        this->m_wifi_stationCount = _m_wifi_stationCount;
        this->apCsmaNodeContainer.Create(m_numberOfNodeINcsmaAPcluster);


        this->csmaHlpr.SetChannelAttribute("DataRate", StringValue("2Gbps")); // TODO:: change data rate
        this->csmaHlpr.SetChannelAttribute("Delay", TimeValue(NanoSeconds(0))); // TODO: change delay time


        this->csmaNetDeviceContainer = csmaHlpr.Install(this->apCsmaNodeContainer); // all the APs will be connected in a CSMA network


        Simulator::sba_INSERT_NODE_GROUP_CHANNEL(this->apCsmaNodeContainer.Get(0)->GetId(), this->groupID, CHANNEL_DUMMY_FOR_THE_AGGREGATOR, AGGREGATOR);

        for (int i = 1; i < this->m_numberOfNodeINcsmaAPcluster; i++)
        {
            this->wifiApNodeContainer.Add(this->apCsmaNodeContainer.Get(i));
        }

        Simulator::sba_INSERT_NODE_GROUP_CHANNEL(this->wifiApNodeContainer.Get(0)->GetId(), this->groupID, CHANNEL_ONE, AP_MODULE);
        Simulator::sba_INSERT_NODE_GROUP_CHANNEL(this->wifiApNodeContainer.Get(1)->GetId(), this->groupID, CHANNEL_TWO, AP_MODULE);
        Simulator::sba_INSERT_NODE_GROUP_CHANNEL(this->wifiApNodeContainer.Get(2)->GetId(), this->groupID, CHANNEL_THREE, AP_MODULE);

        this->wifiStaNodeContainer.Create(this->m_wifi_stationCount);
        for (int i = 0; i < this->m_wifi_stationCount; i++)
        {
            Simulator::sba_INSERT_NODE_GROUP_CHANNEL(this->wifiStaNodeContainer.Get(i)->GetId(), this->groupID, this->defaultChannel, NORMAL_NODE);
        }

        this->apDevicesContainer = wifiHlpr.Install(wifiPhyHlpr, wifiMacHlpr, this->wifiApNodeContainer);
        this->staDevicesContainer = wifiHlpr.Install(wifiPhyHlpr, wifiMacHlpr, this->wifiStaNodeContainer);

#if 1
        MobilityHelper mobility;
        Ptr<ListPositionAllocator> positionAlloc = CreateObject<ListPositionAllocator>();
        positionAlloc->Add(Vector(0.0, 0.0, 0.0));
        mobility.SetPositionAllocator(positionAlloc);
        mobility.SetMobilityModel("ns3::ConstantPositionMobilityModel");
        mobility.Install(wifiApNodeContainer);
        mobility.Install(wifiStaNodeContainer);

        for (uint32_t i = 0; i < wifiApNodeContainer.GetN(); i++)
        {
            Ptr<MobilityModel> mobility = (wifiApNodeContainer.Get(i))->GetObject<MobilityModel>();
            Vector pos;
            pos.x = this->ap_x;
            pos.y = this->ap_y;
            pos.z = 0;
            mobility->SetPosition(pos);
        }

        for (uint32_t i = 0; i < wifiStaNodeContainer.GetN(); i++)
        {
            Ptr<MobilityModel> mobility = (wifiStaNodeContainer.Get(i))->GetObject<MobilityModel>();

            double r = this->rand->GetInteger(0, this->ap_txRadious);
            double theta_deg = this->rand->GetInteger(0, 360);
            double theta_rad = theta_deg*(3.141592654/180.0);
            int x = r*std::cos(theta_rad);
            int y = r*std::sin(theta_rad);
            int relative_x = x + this->ap_x;
            int relative_y = y + this->ap_y;

            Vector pos;
            pos.x = relative_x;
            pos.y = relative_y;
            pos.z = 0;
            mobility->SetPosition(pos);

        }
#else         
        /*Mobility  model start*/
        MobilityHelper mobility;

        mobility.SetPositionAllocator("ns3::GridPositionAllocator",
                "MinX", DoubleValue(0.0),
                "MinY", DoubleValue(0.0),
                "DeltaX", DoubleValue(10.0),
                "DeltaY", DoubleValue(10.0),
                "GridWidth", UintegerValue(3),
                "LayoutType", StringValue("RowFirst"));

        /*
        mobility.SetMobilityModel("ns3::RandomWalk2dMobilityModel",
                "Bounds", RectangleValue(Rectangle(-50, 50, -50, 50)));
        mobility.Install(wifiStaNodes);
         */

        mobility.SetMobilityModel("ns3::ConstantPositionMobilityModel");
        mobility.Install(wifiApNodeContainer);
        mobility.Install(wifiStaNodeContainer);
        /*Mobility model end*/
#endif
        g_internetStkHlpr.Install(apCsmaNodeContainer);
        g_internetStkHlpr.Install(wifiStaNodeContainer);

        /*
        std::string part1 = "10.";
        std::string middle = "33";
        std::string lastPart_csma = ".1.0";
        std::string lastPart_wifi = ".2.0";
        */
        char buffer_csma[50];
        sprintf(buffer_csma, "10.%d.1.0",this->groupID);
        
        char buffer_wifi[50];
        sprintf(buffer_wifi, "10.%d.2.0",this->groupID);
        
        
        //ipv4Hlpr.SetBase("10.1.1.0", "255.255.255.0");
        ipv4Hlpr.SetBase(buffer_csma, "255.255.255.0");
        this->csmaIpv4IntfcCntnr = ipv4Hlpr.Assign(csmaNetDeviceContainer);


        //ipv4Hlpr.SetBase("10.1.2.0", "255.255.255.0");
        ipv4Hlpr.SetBase(buffer_wifi, "255.255.255.0");
        this->apInterfaceContainer = ipv4Hlpr.Assign(apDevicesContainer);
        this->staInterfaceContainer = ipv4Hlpr.Assign(staDevicesContainer);

        //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
        int wifiNodeID;
        int csmaNodeID;
        Mac48Address wifiMAC;
        Mac48Address csmaMAC;
        Ipv4Address wifiIP;
        Ipv4Address csmaIP;
        int wifiIndex = 0;
        int csmaIndex = 1 + wifiIndex;

        
        wifiNodeID = apDevicesContainer.Get(wifiIndex)->GetNode()->GetId();
        csmaNodeID = csmaNetDeviceContainer.Get(csmaIndex)->GetNode()->GetId();
        NS_ASSERT(wifiNodeID == csmaNodeID);

        wifiMAC = Mac48Address::ConvertFrom(apDevicesContainer.Get(wifiIndex)->GetAddress());
        csmaMAC = Mac48Address::ConvertFrom(csmaNetDeviceContainer.Get(csmaIndex)->GetAddress());
        wifiIP = this->apInterfaceContainer.GetAddress(wifiIndex);
        csmaIP = this->csmaIpv4IntfcCntnr.GetAddress(csmaIndex);

        Simulator::sba_SET_CHNL_WIFI_Mac(this->groupID, CHANNEL_ONE, wifiMAC, apDevicesContainer.Get(wifiIndex));
        Simulator::sba_SET_CHNL_CSMA_Mac(this->groupID, CHANNEL_ONE, csmaMAC);
        Simulator::sba_SET_CHNL_WIFI_IPAdr(this->groupID, CHANNEL_ONE, wifiIP);
        Simulator::sba_SET_CHNL_CSMA_IPAdr(this->groupID, CHANNEL_ONE, csmaIP);

        //------------------------------------------------------------------------------//

        ++wifiIndex;
        csmaIndex = 1 + wifiIndex;

        wifiNodeID = apDevicesContainer.Get(wifiIndex)->GetNode()->GetId();
        csmaNodeID = csmaNetDeviceContainer.Get(csmaIndex)->GetNode()->GetId();
        NS_ASSERT(wifiNodeID == csmaNodeID);

        wifiMAC = Mac48Address::ConvertFrom(apDevicesContainer.Get(wifiIndex)->GetAddress());
        csmaMAC = Mac48Address::ConvertFrom(csmaNetDeviceContainer.Get(csmaIndex)->GetAddress());
        wifiIP = this->apInterfaceContainer.GetAddress(wifiIndex);
        csmaIP = this->csmaIpv4IntfcCntnr.GetAddress(csmaIndex);

        Simulator::sba_SET_CHNL_WIFI_Mac(this->groupID, CHANNEL_TWO, wifiMAC, apDevicesContainer.Get(wifiIndex));
        Simulator::sba_SET_CHNL_CSMA_Mac(this->groupID, CHANNEL_TWO, csmaMAC);
        Simulator::sba_SET_CHNL_WIFI_IPAdr(this->groupID, CHANNEL_TWO, wifiIP);
        Simulator::sba_SET_CHNL_CSMA_IPAdr(this->groupID, CHANNEL_TWO, csmaIP);

        //------------------------------------------------------------------------------//

        ++wifiIndex;
        csmaIndex = 1 + wifiIndex;

        wifiNodeID = apDevicesContainer.Get(wifiIndex)->GetNode()->GetId();
        csmaNodeID = csmaNetDeviceContainer.Get(csmaIndex)->GetNode()->GetId();
        NS_ASSERT(wifiNodeID == csmaNodeID);

        wifiMAC = Mac48Address::ConvertFrom(apDevicesContainer.Get(wifiIndex)->GetAddress());
        csmaMAC = Mac48Address::ConvertFrom(csmaNetDeviceContainer.Get(csmaIndex)->GetAddress());
        wifiIP = this->apInterfaceContainer.GetAddress(wifiIndex);
        csmaIP = this->csmaIpv4IntfcCntnr.GetAddress(csmaIndex);

        Simulator::sba_SET_CHNL_WIFI_Mac(this->groupID, CHANNEL_THREE, wifiMAC, apDevicesContainer.Get(wifiIndex));
        Simulator::sba_SET_CHNL_CSMA_Mac(this->groupID, CHANNEL_THREE, csmaMAC);
        Simulator::sba_SET_CHNL_WIFI_IPAdr(this->groupID, CHANNEL_THREE, wifiIP);
        Simulator::sba_SET_CHNL_CSMA_IPAdr(this->groupID, CHANNEL_THREE, csmaIP);

        //------------------------------------------------------------------------//
        sba_print_Chnl_Mac_Relation(Simulator::sba_get_group_chnl_mac_WIFI_MAPPING(), Simulator::sba_get_group_chnl_mac_CSMA_MAPPING());
        sba_print_Chnl_IPAdr_Relation(Simulator::sba_get_wifiGroupBased_IPAdr_ChnlMap(), Simulator::sba_get_csmaGroupBased_IPAdr_ChnlMap());

        //////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////



        NS_LOG_UNCOND("=============================================================");
        for (int i = 0; i < (this->m_numberOfNodeINcsmaAPcluster - 1); i++)
        {
            Address macAdr = apDevicesContainer.Get(i)->GetAddress();
            Mac48Address mac48Adr = Mac48Address::ConvertFrom(macAdr);
            Ipv4Address ipAdr = this->apInterfaceContainer.GetAddress(i);
            uint32_t nodeID = apDevicesContainer.Get(i)->GetNode()->GetId();

            NS_LOG_UNCOND(i << " > node " << nodeID << "; wifi MAC " << macAdr << ", mac 48 " << mac48Adr << "; IP " << ipAdr);
            Simulator::sba_insert_nodeID_AddrType_MacADR_IPAdr_MAP(nodeID, WIFI_ADDR, mac48Adr, ipAdr);
        }
        NS_LOG_UNCOND("=============================================================");
        for (uint32_t i = 0; i< this->csmaNetDeviceContainer.GetN(); i++)
        {
            Address macAdr = csmaNetDeviceContainer.Get(i)->GetAddress();
            Mac48Address mac48Adr = Mac48Address::ConvertFrom(macAdr);
            Ipv4Address ipAdr = this->csmaIpv4IntfcCntnr.GetAddress(i);
            uint32_t nodeID = csmaNetDeviceContainer.Get(i)->GetNode()->GetId();

            NS_LOG_UNCOND(i << " > node " << nodeID << "; csma MAC " << macAdr << ", mac 48 " << mac48Adr << "; IP " << ipAdr);
            Simulator::sba_insert_nodeID_AddrType_MacADR_IPAdr_MAP(nodeID, CSMA_ADDR, mac48Adr, ipAdr);
        }
        NS_LOG_UNCOND("=============================================================");
        for (uint32_t i = 0; i< this->staDevicesContainer.GetN(); i++)
        {
            Address macAdr = staDevicesContainer.Get(i)->GetAddress();
            Mac48Address mac48Adr = Mac48Address::ConvertFrom(macAdr);
            Ipv4Address ipAdr = this->staInterfaceContainer.GetAddress(i);
            uint32_t nodeID = staDevicesContainer.Get(i)->GetNode()->GetId();
            NS_LOG_UNCOND(i << " > node " << nodeID << "; station MAC " << macAdr << ", mac 48 " << mac48Adr << "; IP " << ipAdr);
            Simulator::sba_insert_nodeID_AddrType_MacADR_IPAdr_MAP(nodeID, WIFI_ADDR, mac48Adr, ipAdr);
        }
        NS_LOG_UNCOND("=============================================================");

        /* TCP Sink start*/
        uint16_t TCPsinkPort = 8080;
        this->apTCPsinkAddress = InetSocketAddress(csmaIpv4IntfcCntnr.GetAddress(0), TCPsinkPort); // interface of sink
        PacketSinkHelper packetSinkHelperTCP("ns3::TcpSocketFactory", InetSocketAddress(Ipv4Address::GetAny(), TCPsinkPort));

        uint32_t nodeID = apCsmaNodeContainer.Get(0)->GetId();
        ApplicationContainer apTCPsinkApp = packetSinkHelperTCP.Install(apCsmaNodeContainer.Get(0));
        this->insert_nodeID_SinkType_AppContainer(nodeID, TCP_TYPE, apTCPsinkApp);
        this->getAppContainer_givenNodeID_SinkType(nodeID, TCP_TYPE).Start(Seconds(0.0));
        this->getAppContainer_givenNodeID_SinkType(nodeID, TCP_TYPE).Stop(Seconds(this->simulationTimeSec));

        /* TCP Sink end*/

        /* UDP sink start */
        uint16_t udpSinkPort = 1010;
        this->apUDPsinkAddress = InetSocketAddress(this->csmaIpv4IntfcCntnr.GetAddress(0), udpSinkPort); // interface of sink
        PacketSinkHelper packetSinkHelperUDP("ns3::UdpSocketFactory", InetSocketAddress(Ipv4Address::GetAny(), udpSinkPort));

        ApplicationContainer apUDPsinkApp = packetSinkHelperUDP.Install(apCsmaNodeContainer.Get(0));
        this->insert_nodeID_SinkType_AppContainer(nodeID, UDP_TYPE, apUDPsinkApp);
        this->getAppContainer_givenNodeID_SinkType(nodeID, UDP_TYPE).Start(Seconds(0.0));
        this->getAppContainer_givenNodeID_SinkType(nodeID, UDP_TYPE).Stop(Seconds(this->simulationTimeSec));
        /* UDP sink end*/

    }

    void startTCPfromStaToAP(int _staID, uint32_t _pktSize, uint32_t _pktNumber, double startSecond, double endSecond, DataRate dataRate)
    {
        uint32_t TCPpacketSize = _pktSize; // bytes
        uint32_t TCPnumPackets = _pktNumber;
        int staID = _staID;

        NS_ASSERT_MSG((0 <= staID && staID < m_wifi_stationCount), "wifi stationID index out of bound !");

        /*TCP Source start*/
        Ptr<Socket> ns3TcpSocket = Socket::CreateSocket(wifiStaNodeContainer.Get(staID), TcpSocketFactory::GetTypeId());
        // Trace Congestion window
        ns3TcpSocket->TraceConnectWithoutContext("CongestionWindow", MakeCallback(&CwndChange));


        // Create TCP application at n0
        Ptr<MyApp> app = CreateObject<MyApp> ();
        app->Setup(ns3TcpSocket, this->apTCPsinkAddress, TCPpacketSize, TCPnumPackets, dataRate); //DataRate("250Kbps")
        wifiStaNodeContainer.Get(staID)->AddApplication(app);
        //Config::ConnectWithoutContext("/NodeList/*/ApplicationList/*/$ns3::PacketSink/Rx", MakeCallback(&RcvBySink));
        app->SetStartTime(Seconds(startSecond));
        app->SetStopTime(Seconds(endSecond));
    }

    void startTCPfromAPtoSTA(int _staID, uint32_t _pktSize, uint32_t _pktNumber, double startSecond, double endSecond, DataRate dataRate)
    {
        uint32_t TCPpacketSize = _pktSize; // bytes
        uint32_t TCPnumPackets = _pktNumber;
        int staID = _staID;

        NS_ASSERT_MSG((0 <= staID && staID < m_wifi_stationCount), "wifi stationID index out of bound !");

        uint16_t sinkPort = 8080;
        Address staSinkAddress;
        staSinkAddress = InetSocketAddress(this->staInterfaceContainer.GetAddress(staID), sinkPort); // interface of sink
        PacketSinkHelper packetSinkHelperTCP("ns3::TcpSocketFactory", InetSocketAddress(Ipv4Address::GetAny(), sinkPort));

        uint32_t nodeID = wifiStaNodeContainer.Get(staID)->GetId();
        ApplicationContainer sinkApp = packetSinkHelperTCP.Install(wifiStaNodeContainer.Get(staID));
        this->insert_nodeID_SinkType_AppContainer(nodeID, TCP_TYPE, sinkApp);
        this->getAppContainer_givenNodeID_SinkType(nodeID, TCP_TYPE).Start(Seconds(0.0)); // turn on the SINK for all the time
        this->getAppContainer_givenNodeID_SinkType(nodeID, TCP_TYPE).Stop(Seconds(this->simulationTimeSec)); // turn on the SINK for all the time

        Ptr<Socket> ns3TcpSocket = Socket::CreateSocket(apCsmaNodeContainer.Get(0), TcpSocketFactory::GetTypeId());
        // Trace Congestion window
        ns3TcpSocket->TraceConnectWithoutContext("CongestionWindow", MakeCallback(&CwndChange));


        // Create TCP application at n0
        Ptr<MyApp> app = CreateObject<MyApp> ();
        app->Setup(ns3TcpSocket, staSinkAddress, TCPpacketSize, TCPnumPackets, dataRate);
        apCsmaNodeContainer.Get(0)->AddApplication(app);
        //Config::ConnectWithoutContext("/NodeList/*/ApplicationList/*/$ns3::PacketSink/Rx", MakeCallback(&RcvBySink));
        app->SetStartTime(Seconds(startSecond));
        app->SetStopTime(Seconds(endSecond));

    }

    void startUDPfromSTAtoAP(int _staID, uint32_t _pktSize, uint32_t _pktNumber, double startSecond, double endSecond, DataRate dataRate)
    {
        int staID = _staID;
        NS_ASSERT_MSG((0 <= staID && staID < m_wifi_stationCount), "wifi stationID index out of bound !");

        uint32_t UDPpacketSize = _pktSize; // bytes
        uint32_t UDPnumPackets = _pktNumber;

        Ptr<Socket> ns3UdpSocket = Socket::CreateSocket(wifiStaNodeContainer.Get(staID), UdpSocketFactory::GetTypeId());

        // Create TCP application at n0
        Ptr<MyApp> app = CreateObject<MyApp> ();
        app->Setup(ns3UdpSocket, this->apUDPsinkAddress, UDPpacketSize, UDPnumPackets, dataRate);
        wifiStaNodeContainer.Get(staID)->AddApplication(app);
        app->SetStartTime(Seconds(startSecond));
        app->SetStopTime(Seconds(endSecond));

    }

    void startUDPfromAPtoSTA(int _staID, uint32_t _pktSize, uint32_t _pktNumber, double startSecond, double endSecond, DataRate dataRate)
    {
        int staID = _staID;
        NS_ASSERT_MSG((0 <= staID && staID < m_wifi_stationCount), "wifi stationID index out of bound !");

        uint16_t UDPsinkPort = 1010;
        uint32_t UDPpacketSize = _pktSize; // bytes
        uint32_t UDPnumPackets = _pktNumber;

        Address staUDPsinkAddress;
        staUDPsinkAddress = InetSocketAddress(this->staInterfaceContainer.GetAddress(staID), UDPsinkPort); // interface of sink
        PacketSinkHelper packetSinkHelperUDP("ns3::UdpSocketFactory", InetSocketAddress(Ipv4Address::GetAny(), UDPsinkPort));

        uint32_t nodeID = wifiStaNodeContainer.Get(staID)->GetId();
        ApplicationContainer sinkApp = packetSinkHelperUDP.Install(wifiStaNodeContainer.Get(staID));
        this->insert_nodeID_SinkType_AppContainer(nodeID, UDP_TYPE, sinkApp);
        this->getAppContainer_givenNodeID_SinkType(nodeID, UDP_TYPE).Start(Seconds(0.0)); // turn on the SINK for all the time
        this->getAppContainer_givenNodeID_SinkType(nodeID, UDP_TYPE).Stop(Seconds(this->simulationTimeSec)); // turn on the SINK for all the time

        Ptr<Socket> ns3UdpSocket = Socket::CreateSocket(apCsmaNodeContainer.Get(0), UdpSocketFactory::GetTypeId());

        // Create TCP application at n0
        Ptr<MyApp> app = CreateObject<MyApp> ();
        app->Setup(ns3UdpSocket, staUDPsinkAddress, UDPpacketSize, UDPnumPackets, dataRate);
        apCsmaNodeContainer.Get(0)->AddApplication(app);
        app->SetStartTime(Seconds(startSecond));
        app->SetStopTime(Seconds(endSecond));

    }

};

int main(int argc, char *argv[])
{
    //Simulator::sba_setSimulationType(SIMULATION_MULTI_RADIO_MULTI_CHANNEL); // WARNIGN :: make it as the first line
    Simulator::sba_setSimulationType(SIMULATION_MULTI_RADIO_SINGLE_CHANNEL);  // WARNING:: make it as the first line

#if 0
    LogComponentEnable("CsmaNetDevice", LOG_LEVEL_ALL);
    LogComponentEnable("CsmaNetDevice", LOG_PREFIX_FUNC);
    LogComponentEnable("CsmaNetDevice", LOG_PREFIX_NODE);
    LogComponentEnable("CsmaNetDevice", LOG_PREFIX_TIME);
#endif 

    CommandLine cmd;
    cmd.Parse(argc, argv);

    double simulationTime = 2.0;

    bool verbose = false;

    //RngSeedManager::SetSeed(3); // Changes seed from default of 1 to 3
    //RngSeedManager::SetRun(7); // Changes run number from default of 1 to 7
    // Now, create random variables
    Ptr<UniformRandomVariable> rand = CreateObject<UniformRandomVariable> ();
    
  
    /////////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////////
/*
    // Set to a non-QoS upper mac
    NqosWifiMacHelper wifiMacHlpr = NqosWifiMacHelper::Default();
    // Set it to adhoc mode
    wifiMacHlpr.SetType("ns3::AdhocWifiMac");
    
    WifiHelper wifiHlpr;
    if (verbose)
    {
        wifiHlpr.EnableLogComponents(); // Turn on all Wifi logging
    }
    
    wifiHlpr.SetStandard(WIFI_PHY_STANDARD_80211a);
    
    // Set Wi-Fi rate manager
    std::string phyMode("OfdmRate54Mbps");
    wifiHlpr.SetRemoteStationManager("ns3::ConstantRateWifiManager",
            "DataMode", StringValue(phyMode),
            "ControlMode", StringValue(phyMode));

    YansWifiPhyHelper wifiPhyHlpr = YansWifiPhyHelper::Default();
    
    YansWifiChannelHelper wifiChnlHlpr;
    wifiChnlHlpr.SetPropagationDelay("ns3::ConstantSpeedPropagationDelayModel");
    wifiChnlHlpr.AddPropagationLoss("ns3::LogDistancePropagationLossModel",
            "Exponent", DoubleValue(3.0));

    InternetStackHelper g_internetStkHlpr;
    Ipv4AddressHelper g_ipv4Hlpr;

    // disable rts cts all the time.
    Config::SetDefault("ns3::WifiRemoteStationManager::RtsCtsThreshold", StringValue("2200"));
    // disable fragmentation
    Config::SetDefault("ns3::WifiRemoteStationManager::FragmentationThreshold", StringValue("2200"));
    
    /////////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////////
    /////////////////////////////////////////////////////////////////////////////////////////
*/

    std::string g_wifiPhyControlMode("ErpOfdmRate54Mbps");
    std::string g_wifiPhyDataMode("ErpOfdmRate54Mbps");

    InternetStackHelper g_internetStkHlpr;
    Ipv4AddressHelper g_ipv4Hlpr;

    // disable rts cts all the time.
    Config::SetDefault("ns3::WifiRemoteStationManager::RtsCtsThreshold", StringValue("2200"));
    // disable fragmentation
    Config::SetDefault("ns3::WifiRemoteStationManager::FragmentationThreshold", StringValue("2200"));

    ////////////////////////////////////////////////////////////////////////////////

    YansWifiPhyHelper wifiPhyHlpr = YansWifiPhyHelper::Default();
    YansWifiChannelHelper wifiChnlHlpr = YansWifiChannelHelper::Default();
    Ptr<YansWifiChannel> chanel = wifiChnlHlpr.Create();

    wifiPhyHlpr.SetChannel(chanel);

    WifiHelper wifiHlpr;
    if (verbose)
    {
        wifiHlpr.EnableLogComponents(); // Turn on all Wifi logging
    }


    wifiHlpr.SetStandard(WIFI_PHY_STANDARD_80211g);

    // Add a non-QoS upper mac, and disable rate control
    NqosWifiMacHelper wifiMacHlpr = NqosWifiMacHelper::Default();
    wifiMacHlpr.SetType("ns3::AdhocWifiMac");
    wifiHlpr.SetRemoteStationManager("ns3::ConstantRateWifiManager",
            "DataMode", StringValue(g_wifiPhyDataMode),
            "ControlMode", StringValue(g_wifiPhyControlMode));




    ////////////////////////////////////////////////////////////////////



    Ptr<AP_STA_Cluster> apStaCluster1 = CreateObject<AP_STA_Cluster> ();
    //apStaCluster->setup(1, CHANNEL_ONE, simulationTime, 1, wifiHlpr, wifiPhyHlpr, wifiMacHlpr, g_ipv4Hlpr, g_internetStkHlpr, 0, 0, 20, rand);
    apStaCluster1->setup(1, CHANNEL_ONE, simulationTime, 1, wifiHlpr, wifiPhyHlpr, wifiMacHlpr, g_ipv4Hlpr, g_internetStkHlpr, 0, 0, 5, rand);
    
    Ptr<AP_STA_Cluster> apStaCluster2 = CreateObject<AP_STA_Cluster> ();
    //apStaCluster2->setup(2, CHANNEL_ONE, simulationTime, 1, wifiHlpr, wifiPhyHlpr, wifiMacHlpr, g_ipv4Hlpr, g_internetStkHlpr, 0, 0, 20, rand);
    apStaCluster2->setup(2, CHANNEL_ONE, simulationTime, 1, wifiHlpr, wifiPhyHlpr, wifiMacHlpr, g_ipv4Hlpr, g_internetStkHlpr, 0, 0, 5, rand);

    
    //apStaCluster->startTCPfromStaToAP(0, 1400, 1, 0.0, simulationTime, DataRate("54Mbps"));
    //apStaCluster->startTCPfromAPtoSTA(0, 1400, 1, 0.0, simulationTime, DataRate("54Mbps"));
    apStaCluster1->startUDPfromAPtoSTA(0, 1400, 100, 0.0, simulationTime, DataRate("54Mbps"));
    //apStaCluster->startUDPfromSTAtoAP(0, 1400, 1, 0.0, simulationTime, DataRate("54Mbps"));
    
    apStaCluster2->startUDPfromAPtoSTA(0, 1400, 100, 0.0, simulationTime, DataRate("54Mbps"));



    //Ipv4GlobalRoutingHelper::PopulateRoutingTables();

#if 1
    Config::ConnectWithoutContext("/NodeList/*/ApplicationList/*/$ns3::PacketSink/Rx", MakeCallback(&RcvBySink));
    Config::ConnectWithoutContext("/NodeList/*/DeviceList/*/$ns3::WifiNetDevice/Phy/PhyTxBegin", MakeCallback(&PhyTxBegin));
    Config::ConnectWithoutContext("/NodeList/*/DeviceList/*/$ns3::WifiNetDevice/Phy/PhyTxEnd", MakeCallback(&PhyTxEnd));
    Config::ConnectWithoutContext("/NodeList/*/DeviceList/*/$ns3::WifiNetDevice/Phy/PhyRxBegin", MakeCallback(&PhyRxBegin));
    Config::ConnectWithoutContext("/NodeList/*/DeviceList/*/$ns3::WifiNetDevice/Phy/PhyRxEnd", MakeCallback(&PhyRxEnd));
    Config::ConnectWithoutContext("/NodeList/*/DeviceList/*/$ns3::WifiNetDevice/Phy/PhyTxDrop", MakeCallback(&PhyTxDrop));
    Config::ConnectWithoutContext("/NodeList/*/DeviceList/*/$ns3::WifiNetDevice/Phy/PhyRxDrop", MakeCallback(&PhyRxDrop));
#endif
    
    wifiPhyHlpr.EnablePcap("Adhoc_4_wifi_NEW_all", apStaCluster1->staDevicesContainer.Get(0), true);
    wifiPhyHlpr.EnablePcap("Adhoc_4_wifi_NEW_", apStaCluster1->staDevicesContainer.Get(0), false);
    apStaCluster1->csmaHlpr.EnablePcapAll("Adhoc_4_csma_NEW_", true);


    
    Simulator::Stop(Seconds(simulationTime));
    Simulator::Run();
    Simulator::Destroy();


    apStaCluster1->throughputInAllSinkSocket();
    apStaCluster2->throughputInAllSinkSocket();

    return 0;


}

/*
 
 bool
    WifiNetDevice::Send(Ptr<Packet> packet, const Address& dest, uint16_t protocolNumber)
    {
        LlcSnapHeader llc;
        llc.SetType(protocolNumber);
        packet->AddHeader(llc);
 
 */